import requests

from typing import Optional, Type, Union, TypeVar, Generic, List, Tuple
from typing.io import IO

from .exc import FurnishError

JSON = Union[int, float, str, dict, list, bool, None]


class Parameter:
    """
    Parameter annotation.
    """

    def __init__(self, type_: Type, name: Optional[str]=None) -> None:
        self.type_ = type_
        self.name = name


class Path(Parameter):
    """
    Path parameter.
    """


class Query(Parameter):
    """
    Query parameter.
    """


class _BaseBody(Parameter):
    """
    Common base class for different request body handlers.
    """


class Body(_BaseBody):
    """
    Request body.
    """


class Json(_BaseBody):
    """
    Request body, serialized as JSON.
    """
    def __init__(self, type_: Type=JSON, name: Optional[str]=None) -> None:
        super().__init__(type_, name)


class File(_BaseBody):
    """
    Multipart request file part.
    """

    def __init__(self, name: Optional[str]=None) -> None:
        super().__init__(Union[IO, Tuple[str, IO]], name)


class Header(Parameter):
    """
    Request header.

    Note the reversed parameters to `__init__`; a header always requires
    a name, but usually will default to a `str`.
    """

    def __init__(self, name: str, type_: Type=str) -> None:
        super().__init__(type_, name)

T = TypeVar("T")


def _deserialize(cls: Type[T], json_: JSON) -> T:
    """
    Simple JSON deserializer.
    """
    if issubclass(cls, list):
        item_cls, = cls.__args__
        obj = [_deserialize(item_cls, x) for x in json_]

    else:
        obj = cls.__new__(cls)
        # TODO: Deserialize attributes
        obj.__dict__.update(json_)

    return obj


# TODO: This adapts requests.Response objects, there should exist
# a response interface that other adapters could implement.
class Response(Generic[T]):
    """
    Wraps actual HTTP client responses.
    """

    def __init__(self,
                 response: requests.Response,
                 body_cls: Optional[Type[T]]):
        self.response = response
        self.body_cls = body_cls
        self._entity = None

    def json(self, **kwgs) -> JSON:
        """
        Returns the body if any, decoded from JSON.
        """
        return self.response.json(**kwgs)

    def _check_error(self):
        try:
            self.response.raise_for_status()

        except requests.HTTPError as e:
            raise FurnishError("an HTTP error occurred") from e

    def body(self) -> T:
        """
        Returns the body deserialized from JSON as `T`.
        """
        if self.body_cls is None:
            raise FurnishError("missing type information")

        if not self._entity:
            self._check_error()
            body_json = self.response.json()
            self._entity = _deserialize(self.body_cls, body_json)

        return self._entity



import pytest
import furnish
from furnish import create, url, get, post, headers, BaseClient, Body, Json

from furnish.exc import FurnishError


@pytest.fixture
def test_headers():
    return {
        "X-Test": "test"
    }


class TestDecorators:

    def test_simple_definition(self):
        class Api:
            pass

        api = create(Api, "")
        assert isinstance(api, BaseClient),\
            "'create' produces 'furnish.BaseClient' instances"

    def test_attributes(self):
        class Api:
            @get("")
            def method1():
                pass

            @get("")
            def method2():
                pass

        api = create(Api, "")
        assert hasattr(api, 'method1') and hasattr(api, 'method1'),\
            "furnish creates methods"

    def test_url(self):
        @url("get", "/items/{id}")
        def fun():
            pass

        assert hasattr(fun, "_furnish"),\
            "'url' decorator sets '_furnish' attribute"

    def test_http_methods(self):
        methods = ["get", "post", "put", "patch", "delete", "head"]
        for method in methods:
            deco = getattr(furnish, method)
            @deco("/")
            def f():
                pass

            assert f._furnish.method == method
            assert f._furnish.template == "/"

    def test_validation(self):
        with pytest.raises(FurnishError,
                           message="Multiple Body parameters raise error"):
            class Api:
                @post("")
                def create(body1: Body(dict), body2: Body(dict)): pass

            create(Api, "")

        with pytest.raises(FurnishError,
                           message="Multiple Json parameters raise error"):
            class Api:
                @post("")
                def create(body1: Json(dict), body2: Json(dict)): pass

            create(Api, "")

        with pytest.raises(FurnishError,
                           message="Missing Path parameters raise error"):
            class Api:
                @get("/{network}/item/{id}/")
                def get():
                    pass

            create(Api, "")

    def test_headers(self, test_headers):
        @headers(test_headers)
        def fun():
            pass

        assert hasattr(fun, "_furnish")
        assert fun._furnish.headers == test_headers

    def test_url_and_headers(self, test_headers):
        @headers(test_headers)
        @url("get", "/")
        def fun():
            pass

        assert hasattr(fun, "_furnish")
        assert fun._furnish.method == "get"
        assert fun._furnish.template == "/"
        assert fun._furnish.headers == test_headers



import pytest
from furnish import create


@pytest.fixture
def api_cls():
    class Api:
        pass

    return Api


class TestClass:

    def test_custom_client(self, api_cls):
        client = object()
        api = create(api_cls, "", client=client)
        assert api.client is client,\
            "'furnish' allows passing custom client"

    def test_base_url(self, api_cls):
        api = create(api_cls, "https://example.org")
        assert api.base_url == "https://example.org"



import pytest
from unittest import mock

from furnish import (
    create, get, post, headers, Path, Query, Body, Json, File, Header,
    Response, FurnishError)

from typing import List

import requests


class Item:

    def __init__(self, foo, bar):
        self.foo = foo
        self.bar = bar


@pytest.fixture(scope="function")
def client():
    return mock.Mock()


@pytest.fixture
def test_headers():
    return {
        "X-Secret": "super secret"
    }


@pytest.fixture
def api_cls(test_headers):
    class Api:

        @get("/")
        def all() -> Response[List[Item]]: pass

        @get("/{id}")
        def item(id: Path(int)) -> Response[Item]: pass

        @post("/")
        def create_item(item: Body(dict)) -> Response: pass

        @post("/")
        def create_item_json(item: Json()) -> Response: pass

        @get("/search")
        def search(q: Query(str)) -> Response[List[Item]]: pass

        @headers(test_headers)
        @get("/secret")
        def secret() -> Response: pass

        @get("/super_secret")
        def super_secret(auth: Header("Authorization")): pass

        @headers(test_headers)
        @get("/combo")
        def combo(auth: Header("Authorization")): pass

        @post("/file")
        def add_file(the_file: File()): pass

    return Api


@pytest.fixture(scope="function")
def api(api_cls, client):
    return create(api_cls, "http://example.org", client=client)


class TestCall:

    def test_get(self, api, client):
        api.all()
        client.assert_called_with("get", "http://example.org/")

        api.item(1)
        client.assert_called_with("get", "http://example.org/1")

        ret = api.search("foo")
        client.assert_called_with("get", "http://example.org/search",
                                  params={ "q": "foo" })
        assert ret.response is client.return_value

    def test_post(self, api, client):
        api.create_item({ "name": "Test" })
        client.assert_called_with("post", "http://example.org/",
                                  data={ "name": "Test" })

    def test_post_json(self, api, client):
        api.create_item_json({ "name": "Test" })
        client.assert_called_with("post", "http://example.org/",
                                  json={ "name": "Test" })

    def test_post_file(self, api, client):
        mock_file = mock.Mock()
        api.add_file(mock_file)
        client.assert_called_with("post", "http://example.org/file",
                                  files={ "the_file": mock_file })

    def test_body(self, api, client):
        the_json = { "foo": 1, "bar": 2 }
        mock_response = mock.Mock()
        mock_response.json.return_value = the_json
        client.return_value = mock_response

        response = api.item(1)
        assert isinstance(response, Response)
        assert response.json() == the_json

        item = response.body()
        assert isinstance(item, Item)
        assert item.foo == 1 and item.bar == 2

    def test_list_body(self, api, client):
        the_json = [
            { "foo": 1, "bar": 2 },
            { "foo": 3, "bar": 4 }
        ]
        mock_response = mock.Mock()
        mock_response.json.return_value = the_json
        client.return_value = mock_response

        response = api.all()
        assert isinstance(response, Response)
        assert response.json() == the_json

        items = response.body()
        assert isinstance(items, list)
        assert isinstance(items[0], Item)
        assert isinstance(items[1], Item)
        assert items[0].foo == 1 and items[0].bar == 2
        assert items[1].foo == 3 and items[1].bar == 4

    def test_request_headers(self, api, client, test_headers):
        api.secret()
        client.assert_called_with("get", "http://example.org/secret",
                                  headers=test_headers)

        api.super_secret("this is auth")
        client.assert_called_with("get", "http://example.org/super_secret",
                                  headers={ "Authorization": "this is auth" })

        combined_headers = { "Authorization": "this is auth" }
        combined_headers.update(test_headers)

        api.combo("this is auth")
        client.assert_called_with("get", "http://example.org/combo",
                                  headers=combined_headers)

    def test_error_response_raises(self, api, client):
        mock_response = mock.Mock()
        mock_response.raise_for_status.side_effect = requests.HTTPError()
        client.return_value = mock_response
        response = api.item(1)
        with pytest.raises(FurnishError):
            response.body()




from typing import Optional
import furnish


class FakeService:

    @furnish.get("/posts")
    def all_posts(user_id: furnish.Query(Optional[int], "userId")=None): pass

    @furnish.post("/posts")
    def create_post(post: furnish.Body(dict)): pass

    @furnish.post("/posts")
    def create_post_json(post: furnish.Json()): pass

    @furnish.get("/posts/{id}")
    def get_post(id: furnish.Path(int)): pass

    @furnish.put("/posts/{id}")
    def replace_post(id: furnish.Path(int),
                     post: furnish.Body(dict)): pass

    @furnish.patch("/posts/{id}")
    def update_post(id: furnish.Path(int),
                    post: furnish.Body(dict)): pass

    @furnish.delete("/posts/{id}")
    def remove_post(id: furnish.Path(int)): pass

    @furnish.get("/comments")
    def comments(post_id: furnish.Query(int, "postId")): pass



from setuptools import setup

with open("README.rst") as f:
    long_description = f.read()

setup(
    name="furnish",
    version="0.2.1",
    description="Create HTTP API clients from Python.",
    long_description=long_description,
    author="Ilja Everilä",
    author_email="saarni@gmail.com",
    license="MIT",
    url="https://github.com/everilae/furnish",
    download_url="https://github.com/everilae/furnish/archive/v0.2.1.tar.gz",
    packages=["furnish"],
    setup_requires=[ "pytest-runner" ],
    tests_require=[ "pytest" ],
    install_requires=[
        "requests",
        "typing",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ]
)



import inspect
import requests
from functools import partial, partialmethod
from collections import namedtuple
from typing import Callable, Optional, Type, Mapping

from .exc import FurnishError
from .types import Path, Query, Body, Json, File, Header, Response
from .utils import LockPick as _LockPick


def _get_parameters(type_, signature):
    return ((name, parameter)
            for name, parameter in signature.parameters.items()
            if isinstance(parameter.annotation, type_))


def _check_signature(signature, template):
    """
    Check that the given `Signature` is valid.
    """
    pick = _LockPick()
    template.format_map(pick)
    path_vars = {name for name, _ in _get_parameters(Path, signature)}
    path_vars_diff = pick.keys - path_vars

    if path_vars_diff:
        raise FurnishError(
            "missing Path parameters: {}".format(path_vars_diff))

    for type_ in [Body, Json]:
        if len(list(_get_parameters(type_, signature))) > 1:
            raise FurnishError(
                "multiple parameters annotated as {}".format(type_.__name__))


def _get_args(type_, signature, bound):
    v = {}
    for name, parameter in _get_parameters(type_, signature):
        value = bound.arguments[name]
        v[parameter.annotation.name or name] = value

    return v


def _get_arg(type_, signature, bound):
    args = _get_args(type_, signature, bound)

    if not args:
        return

    arg, = args.values()
    return arg

_path = partial(_get_args, Path)
_query = partial(_get_args, Query)
_headers = partial(_get_args, Header)
_body = partial(_get_arg, Body)
_json = partial(_get_arg, Json)
_files = partial(_get_args, File)


def _response_cls(signature):
    cls = signature.return_annotation
    if cls is inspect.Signature.empty:
        return None

    return cls


def _body_cls(response_cls):
    body_cls = None
    # Python 3.8 generics don't have __args__ unless specialized.
    args = getattr(response_cls, "__args__", None)

    if args:
        body_cls = args[0]

    return body_cls


_default_client = requests.request


class BaseClient:
    """
    Base client class.
    """

    def __init__(self,
                 base_url: str,
                 client=_default_client,
                 response_cls: Type[Response]=Response):
        self.base_url = base_url
        self.client = client
        self.response_cls = response_cls

    def _call(self,
              signature: inspect.Signature,
              method: str,
              template: str,
              headers: Optional[Mapping[str, str]],
              response_cls: Optional[Type[Response]],
              *args, **kwgs) -> Response:
        """
        Send an HTTP request using `client`.
        """
        bound = signature.bind(*args, **kwgs)
        bound.apply_defaults()

        url = self.base_url + template.format_map(_path(signature, bound))
        kwgs = {}

        query = _query(signature, bound)
        if query:
            kwgs["params"] = query

        body = _body(signature, bound)
        if body:
            kwgs["data"] = body

        json_ = _json(signature, bound)
        if json_:
            kwgs["json"] = json_

        files = _files(signature, bound)
        if files:
            kwgs["files"] = files

        combined_headers = _headers(signature, bound)
        if headers:
            combined_headers.update(headers)

        if combined_headers:
            kwgs["headers"] = combined_headers

        response = self.client(method, url, **kwgs)

        response_cls = response_cls or self.response_cls
        body_cls = _body_cls(response_cls)
        return response_cls(response, body_cls)


def _isfurnished(member):
    return inspect.isfunction(member) and hasattr(member, "_furnish")


def create(cls: Type,
           *args,
           base_cls: Type[BaseClient]=BaseClient,
           **kwgs) -> Type[BaseClient]:
    """
    Create HTTP API client from class definition.
    """
    namespace = {}
    for name, member in inspect.getmembers(cls, _isfurnished):
        method, template, headers = member._furnish

        if method is None:
            raise FurnishError(
                "method not defined – possibly missing url decorator?")

        if template is None:
            raise FurnishError(
                "template not defined – possibly missing url decorator?")

        signature = inspect.signature(member)
        _check_signature(signature, template)

        response_cls = _response_cls(signature)
        namespace[name] = partialmethod(
            base_cls._call, signature, method, template, headers, response_cls)

    return type(cls.__name__, (base_cls,), namespace)(*args, **kwgs)


_Marker = namedtuple("Marker", "method template headers")

Decorator = Callable[[Callable], Callable]


def _mark(f, **kwgs):
    try:
        f._furnish = f._furnish._replace(**kwgs)

    except AttributeError:
        f._furnish = _Marker(*map(kwgs.get, _Marker._fields))


def url(method: str, template: str) -> Decorator:
    """
    Define URL template.
    """
    def decorator(f: Callable) -> Callable:
        _mark(f, method=method, template=template)
        return f

    return decorator

get = partial(url, "get")
post = partial(url, "post")
put = partial(url, "put")
patch = partial(url, "patch")
delete = partial(url, "delete")
head = partial(url, "head")


def headers(headers_map: Mapping[str, str]) -> Decorator:
    """
    Define request headers.
    """
    def decorator(f: Callable) -> Callable:
        _mark(f, headers=headers_map)
        return f

    return decorator
